------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--                   'K e r n e l . F i l e _ S y s t e m'
--
--                                    Body
--
--
--  File 'k-file_system.adb'                                    By Fguerreira.
--
--
--  File system I/O functions
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------

with Kernel.File_System_Data_Types;   use Kernel.File_System_Data_Types;

with Kernel.Devices_Table;  use Kernel.Devices_Table;

with Kernel.Tasks_Operations.Internals;
with Kernel.Mutexes;  use Kernel.Mutexes;
with Kernel.Mutexes.Internals;
with General_Constants;
with Ada.Unchecked_Conversion;
with System;

package body Kernel.File_System is

   package GC renames General_Constants;

   use type Int;

   -----------
   --  EOF  --
   -----------
   EOF : constant := -1;
   --  Any change in this value should be also done in 'stdio.h'.

   package TOI renames Kernel.Tasks_Operations.Internals;

   ---------------------------------------------------------------------------
   --  File operations  -------------------------------------------------------
   ---------------------------------------------------------------------------

   ----------
   -- Open --
   ----------
   function Open
     (Path_Name : access Path;
      Mode : in File_Access_Mode)
     return Int is

      procedure Search_File (Path_In : in Path;
                             Device_Found : out Boolean;
                             The_File_Number : out Device_File_Number;
                             Create_File     : in Boolean);

      Flags : Integer;
      Internal_Procedure_Error : Int;
      The_Fd : File_Descriptor;
      The_Device_File_Number : Device_File_Number;
      Device_File_Found : Boolean := False;
      Fd_Found : Boolean := False;

      --  Search file, also write the name of a new created file in
      --  'Kernel.Devices_Table.The_Device_Files_Table'
      procedure Search_File  (Path_In         : in Path;
                              Device_Found    : out Boolean;
                              The_File_Number : out Device_File_Number;
                              Create_File     : in Boolean) is
      begin
         Device_Found := False;
         for I in Device_File_Number'Range loop
            if The_Device_Files_Table (I).Device_Used then
               --  Compare Paths
               for J in Path'Range loop
                  if Create_File then
                     --  Creating a file in a directory ("/xxx/ ")
                     if (The_Device_Files_Table (I).File_Name (J) /=
                         Path_In (J)) then
                        if (The_Device_Files_Table (I).File_Name (J - 1 .. J) =
                            "/ ") then
                           for K in Path'Range loop
                              exit when Path_In (K) = ASCII.NUL;
                              The_Device_Files_Table (I).File_Name (K) :=
                                Path_In (K);
                           end loop;
                           The_File_Number := I;
                           Device_Found := True;
                           return;
                        else
                           exit;
                        end if;
                     end if;
                  else
                     --  Looking for a file (minimum path=2chars --> "/a")
                     if (The_Device_Files_Table (I).File_Name (J) = ' ' and
                         (J > 2)) then
                        The_File_Number := I;
                        Device_Found := True;
                        return;
                     elsif (The_Device_Files_Table (I).File_Name (J) /=
                            Path_In (J)) then
                        exit;
                     end if;
                  end if;
               end loop;
               --  End Compare Paths
            end if;
         end loop;
      end Search_File;

      use type Kernel.File_System_Data_Types.File_Access_Mode;

   begin

      Kernel.Enter_Critic_Section (Flags);

      TOI.Reset_POSIX_Error;

      if (((Mode and File_Access_Mode (GC.FILE_MODE_MASK)) /= READ_ONLY) and
          ((Mode and File_Access_Mode (GC.FILE_MODE_MASK)) /= WRITE_ONLY) and
          ((Mode and File_Access_Mode (GC.FILE_MODE_MASK)) /= READ_WRITE)) then
         TOI.Set_POSIX_Error_And_Leave_Critic_Section (INVALID_ARGUMENT,
                                                       Flags);
         return -1;
      end if;

      Search_File (Path_Name.all, Device_File_Found,
                   The_Device_File_Number, False);
      if ((not Device_File_Found) and
          ((Mode and File_Access_Mode (O_CREAT)) = O_CREAT)) then
         Search_File (Path_Name.all, Device_File_Found,
                      The_Device_File_Number, True);
      end if;

      if not Device_File_Found then
         TOI.Set_POSIX_Error_And_Leave_Critic_Section
           (NO_SUCH_FILE_OR_DIRECTORY, Flags);
         return -1;
      end if;

      for I in File_Descriptor loop
         if not The_Fd_Table (I).Fd_Used and not Fd_Found then
            The_Fd_Table (I).File_Open_Status     :=
              Kernel.File_System_Data_Types.File_Access_Mode (Mode);
            The_Fd_Table (I).Device_File_Assigned :=
              The_Device_File_Number;
            The_Fd_Table (I).Fd_Used := True;
            Fd_Found := True;
            The_Fd := I;
         end if;
      end loop;

      if not Fd_Found then
         TOI.Set_POSIX_Error_And_Leave_Critic_Section (TOO_MANY_OPEN_FILES,
                                                       Flags);
         return -1;
      end if;

      Kernel.Leave_Critic_Section (Flags);

      if The_Driver_Table (The_Device_Files_Table
                           (The_Device_File_Number).Major_Number).Open /= null
      then
         Internal_Procedure_Error := The_Driver_Table (The_Device_Files_Table
                           (The_Device_File_Number).Major_Number).Open.all
           (The_Fd, Kernel.File_System_Data_Types.File_Access_Mode (Mode));

         if Internal_Procedure_Error = -1 then
            Kernel.Enter_Critic_Section (Flags);
            The_Fd_Table (The_Fd).Fd_Used := False;
            Kernel.Leave_Critic_Section (Flags);
            return -1;
         end if;

      end if;

      return Int (The_Fd);
   end Open;

   --  ***********************  --
   --  File_Descriptor_Correct  --
   --  ***********************  --

   function File_Descriptor_Correct (File_To_Check : in File_Descriptor)
                                    return Boolean is
   begin
      if ((Int (File_To_Check) < Int (File_Descriptor'First) or
           Int (File_To_Check) > Int (File_Descriptor'Last)) or else
          not The_Fd_Table (File_To_Check).Fd_Used) then
         TOI.Set_POSIX_Error (BAD_FILE_DESCRIPTOR);
         return False;
      end if;
      return True;
   end File_Descriptor_Correct;
   pragma Inline (File_Descriptor_Correct);

   -----------
   -- Close --
   -----------
   function Close (Fd : in File_Descriptor)
                  return Int is

      Flags : Integer;
      Internal_Procedure_Error : Int := 0;
      The_Device_File_Number : Device_File_Number;
      The_Major : Major;

   begin
      Kernel.Enter_Critic_Section (Flags);

      TOI.Reset_POSIX_Error;

      if not File_Descriptor_Correct (Fd) then
         Kernel.Leave_Critic_Section (Flags);
         return -1;
      end if;

      The_Device_File_Number := The_Fd_Table (Fd).Device_File_Assigned;

      The_Major :=
        The_Device_Files_Table (The_Device_File_Number).Major_Number;

      Kernel.Leave_Critic_Section (Flags);

      if The_Driver_Table (The_Major).Close /= null then
         Internal_Procedure_Error :=
           The_Driver_Table (The_Major).Close.all (Fd);
      end if;

      Kernel.Enter_Critic_Section (Flags);
      --  Relases the Fd table slot
      The_Fd_Table (Fd).Fd_Used := False;
      Kernel.Leave_Critic_Section (Flags);

      return Internal_Procedure_Error;

   end Close;

   ----------
   -- Read --
   ----------
   function Read
     (Fd            : in File_Descriptor;
      Buffer_Ptr    : in Buffer_Ac;
      Bytes_To_Read : in Buffer_Length)
     return Int is

      Flags      : Integer;
      Bytes_Read : Int := 0;
      The_Device_File_Number : Device_File_Number;
      The_Major : Major;
      function To_Buffer_Ac is new Ada.Unchecked_Conversion (System.Address,
                                                             Buffer_Ac);

   begin

      Kernel.Enter_Critic_Section (Flags);

      TOI.Reset_POSIX_Error;

      if not File_Descriptor_Correct (Fd) then
         Kernel.Leave_Critic_Section (Flags);
         return -1;
      end if;

      if (The_Fd_Table (Fd).File_Open_Status and READ_ONLY) = 0 then
         TOI.Set_POSIX_Error_And_Leave_Critic_Section
           (PERMISSION_DENIED, Flags);
         return -1;
      elsif Bytes_To_Read = 0 then
         Kernel.Leave_Critic_Section (Flags);
         return 0;
      end if;

      The_Device_File_Number := The_Fd_Table (Fd).Device_File_Assigned;

      The_Major :=
        The_Device_Files_Table (The_Device_File_Number).Major_Number;

      Kernel.Leave_Critic_Section (Flags);

      if The_Driver_Table (The_Major).Read /= null then
         if The_Fd_Table (Fd).Ungot_Char_Pending then
            --  There is a char ungot.
            Buffer_Ptr (Buffer_Ptr'First) := The_Fd_Table (Fd).Ungot_Char;
            The_Fd_Table (Fd).Ungot_Char_Pending := False;
            Bytes_Read :=
              The_Driver_Table (The_Major).Read.all
              (Fd,
               To_Buffer_Ac (Buffer_Ptr (Buffer_Ptr'First + 1)'Address),
               Bytes_To_Read - 1) + 1;
         else
            Bytes_Read :=
              The_Driver_Table (The_Major).Read.all (Fd,
                                                     Buffer_Ptr,
                                                     Bytes_To_Read);
         end if;
      end if;

      return Bytes_Read;
   end Read;

   -----------
   -- Write --
   -----------
   function Write
     (Fd             : in File_Descriptor;
      Buffer_Ptr     : in Buffer_Ac;
      Bytes_To_Write : in Buffer_Length)
     return Int is

      Flags         : Integer;
      Bytes_Written : Int := 0;
      The_Device_File_Number : Device_File_Number;
      The_Major : Major;

   begin

      Kernel.Enter_Critic_Section (Flags);

      TOI.Reset_POSIX_Error;

      if not File_Descriptor_Correct (Fd) then
         Kernel.Leave_Critic_Section (Flags);
         return -1;
      end if;

      if (The_Fd_Table (Fd).File_Open_Status and WRITE_ONLY) = 0 then
         TOI.Set_POSIX_Error_And_Leave_Critic_Section
           (PERMISSION_DENIED, Flags);
         return -1;
      elsif Bytes_To_Write = 0 then
         Kernel.Leave_Critic_Section (Flags);
         return 0;
      end if;

      The_Device_File_Number := The_Fd_Table (Fd).Device_File_Assigned;

      The_Major :=
        The_Device_Files_Table (The_Device_File_Number).Major_Number;

      Kernel.Leave_Critic_Section (Flags);

      if The_Driver_Table (The_Major).Write /= null then
         Bytes_Written :=
           The_Driver_Table (The_Major).Write.all (Fd, Buffer_Ptr,
                                                   Bytes_To_Write);

         if Bytes_Written < 1 then
            TOI.Set_POSIX_Error (NOT_ENOUGH_SPACE);
         end if;
      end if;

      return Bytes_Written;
   end Write;

   ------------
   -- Ungetc --
   ------------
   function Ungetc (Ungot_Char : in Int;
                    Fd         : in File_Descriptor) return Int is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      TOI.Reset_POSIX_Error;

      if not File_Descriptor_Correct (Fd) then
         Kernel.Leave_Critic_Section (Flags);
         return EOF;
      end if;

      Kernel.Leave_Critic_Section (Flags);

      The_Fd_Table (Fd).Ungot_Char := Unsigned_8 (Ungot_Char);
      The_Fd_Table (Fd).Ungot_Char_Pending := True;

      return Ungot_Char;
   end Ungetc;

   -----------
   -- Ioctl --
   -----------
   function Ioctl
     (Fd             : in File_Descriptor;
      Request        : in Ioctl_Option_Value;
      Ioctl_Data_Ptr : in Buffer_Ac)
     return Int is

      Flags : Integer := 0;
      Internal_Procedure_Error : Int := 0;
      The_Device_File_Number : Device_File_Number;
      The_Major : Major;
   begin

      Kernel.Enter_Critic_Section (Flags);

      TOI.Reset_POSIX_Error;

      if not File_Descriptor_Correct (Fd) then
         Kernel.Leave_Critic_Section (Flags);
         return -1;
      end if;

      The_Device_File_Number :=
        The_Fd_Table (Fd).Device_File_Assigned;

      The_Major :=
        The_Device_Files_Table (The_Device_File_Number).Major_Number;

      Kernel.Leave_Critic_Section (Flags);

      if The_Driver_Table (The_Major).Ioctl /= null then
         Internal_Procedure_Error := The_Driver_Table (The_Major).Ioctl.all
           (Fd, Request, Ioctl_Data_Ptr);
      end if;

      return Internal_Procedure_Error;
   end Ioctl;

end Kernel.File_System;
