------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--      'K e r n e l . T i m e d _ H a n d l e r s .
--                                       A d a _ T i m i n g _ E v e n t s'
--
--                                   Body
--
--
--
--  File 'kernel-timed_handlers-ada_timing_events.adb'                 By MAR.
--
--  Support for Ada 2005 standard packages 'Ada.Execution_Time.Timers' and
--  'Ada.Timing_Events'.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with System;
with Ada.Unchecked_Conversion;

with Interfaces.C;

with Hardware_Interface;
with Kernel.Timed_Handlers.Internals;
with Kernel.Timed_Handlers.Operations;
with Configuration_Parameters;
with Kernel.Debug;

package body Kernel.Timed_Handlers.Ada_Timing_Events is

   package THO renames Kernel.Timed_Handlers.Operations;
   package THI renames Kernel.Timed_Handlers.Internals;

   procedure ASSERT (Condition : in Boolean; Msg : in String := "")
                     renames Kernel.Debug.Assert;

   use type Interfaces.C.int;

   type Base_Timing_Event_Ac is access all Base_Timing_Event;

   -------------------------------
   --  Generic_Handler_Wrapper  --
   -------------------------------
   procedure Generic_Handler_Wrapper (Area : in System.Address;
                                      TE   : in TH.Timed_Handler_TE_Ac);
   procedure Generic_Handler_Wrapper (Area : in System.Address;
                                      TE   : in TH.Timed_Handler_TE_Ac) is
      function UC is new
        Ada.Unchecked_Conversion (System.Address, Base_Timing_Event_Ac);
   begin
      if Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (ASSERT (False));
         return;
      end if;

      pragma Debug (ASSERT (not UC (Area).Cleared and
                              UC (Area).Handler /= null));
      UC (Area).Cleared := True;
      UC (Area).Handler (UC (Area).all);  --  Call user's handler
   end Generic_Handler_Wrapper;

   ------------------
   --  Initialize  --
   ------------------
   procedure Initialize (Event : in out Base_Timing_Event) is
      Ret : Interfaces.C.int;
   begin
      if Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (ASSERT (False));
         return;
      end if;

      Event.Cleared := True;
      Event.Handler := null;
      Ret :=
        THO.MaRTE_Timed_Handler_Init (Event.Timed_Handler'Unrestricted_Access,
                                      Event.Timed_Handler.Base_Clock,
                                      Generic_Handler_Wrapper'Access,
                                      Event'Address,
                                      System.Address'Size);
      pragma Debug (ASSERT (Ret = 0));
   end Initialize;

   ----------------
   --  Finalize  --
   ----------------
   procedure Finalize (Event : in out Base_Timing_Event) is
      Ret : Interfaces.C.int;
   begin
      if Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (ASSERT (False));
         return;
      end if;

      Ret := THO.MaRTE_Timed_Handler_Destroy
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (ASSERT (Ret = 0));
      Event.Cleared := True;
      Event.Handler := null;
   end Finalize;

   --------------------------
   --  Set_Handler_HWTime  --
   --------------------------
   procedure Set_Handler_HWTime (Event   : in out Base_Timing_Event;
                                 T       : in HWI.HWTime;
                                 Handler : in Base_Timing_Event_Handler;
                                 Options : in TH.Timed_Handler_Options) is
      Ret : Interfaces.C.int;
      Flags : Integer;
   begin
      if Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (ASSERT (False));
         return;
      end if;
      --  Disable event
      Ret := THO.MaRTE_Timed_Handler_Disable
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (ASSERT (Ret = 0));

      if Handler = null then
         Event.Cleared := True;
         --  Event remains disabled
         return;
      end if;

      if Handler /= Event.Handler then
         --  Handler has changed
         Finalize (Event);
         Initialize (Event);
         Event.Handler := Handler;
      end if;

      Event.Cleared := False;

      Event.Timed_Handler.T := T;
      Event.Timed_Handler.Options := Options;

      Kernel.Enter_Critic_Section (Flags);
      THO.Set (Event.Timed_Handler, HWI.Get_HWTime);
      Kernel.Leave_Critic_Section (Flags);

      --  re-enable event
      Ret := THO.MaRTE_Timed_Handler_Enable
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (ASSERT (Ret = 0));
   end Set_Handler_HWTime;

   -----------------------
   --  Current_Handler  --
   -----------------------
   function Current_Handler (Event : Base_Timing_Event)
                             return Base_Timing_Event_Handler is
   begin
      if Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (ASSERT (False));
         return null;
      end if;

      if Event.Cleared then
         return null;
      else
         return Event.Handler;
      end if;
   end Current_Handler;

   ----------------------
   --  Cancel_Handler  --
   ----------------------
   procedure Cancel_Handler (Event     : in out Base_Timing_Event;
                             Cancelled : out Boolean) is
      Ret : Interfaces.C.int;
   begin
      if Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (ASSERT (False));
         return;
      end if;

      --  Disable event
      Ret := THO.MaRTE_Timed_Handler_Disable
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (ASSERT (Ret = 0));

      Cancelled := not Event.Cleared;
      Event.Cleared := True;
   end Cancel_Handler;

end Kernel.Timed_Handlers.Ada_Timing_Events;
