------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--      'K e r n e l . T i m e d _ H a n d l e r s . O p e r a t i o n s'
--
--                                 Body
--
--
--  File 'kernel-timed_handlers-operations.adb'                        By MAR.
--
--
--
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
-------------------------------------------------------------------------------
with Kernel.Timers; --  For 'CLOCK_REALTIME' and 'CLOCK_MONOTONIC'
with Kernel.Timers.Internals;
with General_Constants;
with Kernel.Timed_Events_And_Timer;
with Hardware_Interface;
with Configuration_Parameters;
with Kernel.Scheduler;
with Kernel.Timed_Handlers.Internals;
with Kernel.Debug;

package body Kernel.Timed_Handlers.Operations is

   package GC    renames General_Constants;
   package TE_T  renames Kernel.Timed_Events_And_Timer;
   package HWI   renames Hardware_Interface;
   package SCHD  renames Kernel.Scheduler;
   package TMRI  renames Kernel.Timers.Internals;
   package CPU_TE_LST       renames Kernel.CPU_Time_Timed_Events_Lists;
   package CPU_TE_LST_ORDER renames Kernel.CPU_Time_Timed_Events_Lists_Order;
   procedure ASSERT (Condition : in Boolean; Msg : in String := "")
     renames Kernel.Debug.Assert;

   --------------------------------
   --  MaRTE_Timed_Handler_Init  --
   --------------------------------
   function MaRTE_Timed_Handler_Init
     (TE_Ac   : in Timed_Handler_TE_Ac;
      Clock   : in Kernel.Timers.Clock_Id;
      Handler : in Handler_Procedure;
      Area    : in System.Address;
      Size    : in Size_T) return Int is

      use type System.Address, Basic_Integer_Types.Size_T;

      Tmp_Timed_Handler_TE : aliased Timed_Handler_Timed_Event;

      procedure Memory_Copy (From  : access Timed_Handler_Timed_Event;
                             To    : Timed_Handler_TE_Ac;
                             Bytes : in     Integer);
      pragma Import (C, Memory_Copy, "memory_copy");

   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;

      if (not TMRI.Clock_Id_OK (Clock) or
          TE_Ac = null or
          (Size > 0 and then Area = System.Null_Address) or
          Handler = null) then
         return INVALID_ARGUMENT;
      end if;
      pragma Debug (ASSERT (TE_Ac.Magic /= INITIALIZED));

      --  To set properly the 'tag' in a timed handler created from a
      --  C program.
      Memory_Copy (From  => Tmp_Timed_Handler_TE'Access,
                   To    => TE_Ac,
                   Bytes => Timed_Handler_Timed_Event'Size / 8);

      TE_Ac.all := (Kernel.CPU_Time_Timed_Events_Lists.Element
                      (TE_Ac.all) with
                    Handler       => Handler,
                    Area          => Area,
                    Options       => 0,
                    Base_Clock    => Clock,
                    Period        => Hardware_Interface.HWTime'Last,
                    Overrun_Count => 0,
                    Active        => False,
                    Enabled       => True,
                    Queued        => False,
                    Pending       => False,
                    Magic         => INITIALIZED);
      return 0;
   end MaRTE_Timed_Handler_Init;

   -----------------------------------
   --  MaRTE_Timed_Handler_Destroy  --
   -----------------------------------
   function MaRTE_Timed_Handler_Destroy (TE_Ac   : in Timed_Handler_TE_Ac)
                                        return Int is
      use type Kernel.Timers.Clock_Id;
      Flags : Integer;
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;
      pragma Debug (ASSERT (TE_Ac /= null  and then
                            TE_Ac.all in Timed_Handler_Timed_Event'Class));

      Kernel.Enter_Critic_Section (Flags);

      if TE_Ac = null or else TE_Ac.Magic /= INITIALIZED then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;
      pragma Debug (ASSERT (TMRI.Clock_Id_OK (TE_Ac.Base_Clock)));

      if TE_Ac.Queued then
         --  Queued event waiting for expiration time: remove from
         --  timed events queue
         if (TE_Ac.Base_Clock = Kernel.Timers.CLOCK_REALTIME or
               TE_Ac.Base_Clock = Kernel.Timers.CLOCK_MONOTONIC) then
            --  Standard event
            SCHD.Remove_Timed_Event (Kernel.Timed_Event_Ac (TE_Ac));
         else
            --  timed handler based on CPU time clock
            CPU_TE_LST.Dequeue
              (CPU_Time_Timed_Event_Ac (TE_Ac),
               TMRI.To_Task_Id (TE_Ac.Base_Clock).CPU_Time_TEs_Q);
         end if;
      elsif TE_Ac.Active then
         --  Expired event whose handler hasn't been executed yet,
         --  because either the event is disabled or handler execution
         --  is disallowed. This situation is detected because the
         --  event is active but is not queued.
         if TE_Ac.Pending then
            --  Remove from the 'Pending_Handlers_List'
            pragma Debug
              (ASSERT (Kernel.Timed_Events_Lists.Is_In_The_List
                       (Kernel.Timed_Event_Ac (TE_Ac),
                        Internals.Pending_Handlers_List)));
            Kernel.Timed_Events_Lists.Dequeue
              (Kernel.Timed_Event_Ac (TE_Ac), Internals.Pending_Handlers_List);
         end if;
         --  For an expired and disabled event nothing special needs
         --  to be done appart to be marked as "NOT_INITIALIZED"
      end if;

      TE_Ac.Magic := NOT_INITIALIZED;

      Kernel.Leave_Critic_Section (Flags);

      return 0;
   end MaRTE_Timed_Handler_Destroy;

   -------------------------------
   --  MaRTE_Timed_Handler_Set  --
   -------------------------------
   function MaRTE_Timed_Handler_Set
     (TE_Ac   : in Timed_Handler_TE_Ac;
      Options : in Basic_Integer_Types.Unsigned_32;
      Time    : in Type_Timespec.Timespec_Ac) return Int is
      Flags : Integer;
      use type Type_Timespec.Timespec_Ac;
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;
      pragma Debug (ASSERT (TE_Ac.all in Timed_Handler_Timed_Event'Class));

      Kernel.Enter_Critic_Section (Flags);

      if (TE_Ac = null
          or else TE_Ac.Magic /= INITIALIZED
          or else Time = null) then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;
      pragma Debug (ASSERT (TMRI.Clock_Id_OK (TE_Ac.Base_Clock)));
      if (TE_Ac.Base_Clock /= Kernel.Timers.CLOCK_REALTIME and
            TE_Ac.Base_Clock /= Kernel.Timers.CLOCK_MONOTONIC and
              (TE_Ac.Options and PERIODIC_HANDLER) /= 0) then
         Kernel.Leave_Critic_Section (Flags);
         --  Periodic timed events based on CPU-time clocks are not supported
         --  yet
         return OPERATION_NOT_SUPPORTED;
      end if;

      TE_Ac.Options := Options;
      TE_Ac.T := Type_Timespec.Timespec_To_HWTime (Time.all);
      TE_Ac.Period := TE_Ac.T; --  Only used in periodic handlers

      --  Set the event
      Set (TE_Ac.all, HWI.Get_HWTime);

      Kernel.Leave_Critic_Section (Flags);

      return 0;
   end MaRTE_Timed_Handler_Set;

   -----------------------------------
   --  MaRTE_Timed_Handler_Disable  --
   -----------------------------------
   function MaRTE_Timed_Handler_Disable (TE_Ac : in Timed_Handler_TE_Ac)
                                        return Int is
      Flags : Integer;
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;
      pragma Debug (ASSERT (TE_Ac.all in Timed_Handler_Timed_Event'Class));

      Kernel.Enter_Critic_Section (Flags);

      if TE_Ac = null or else TE_Ac.Magic /= INITIALIZED then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      TE_Ac.Enabled := False;

      Kernel.Leave_Critic_Section (Flags);

      return 0;
   end MaRTE_Timed_Handler_Disable;

   ----------------------------------
   --  MaRTE_Timed_Handler_Enable  --
   ----------------------------------
   function MaRTE_Timed_Handler_Enable (TE_Ac : in Timed_Handler_TE_Ac)
                                       return Int is
      Flags : Integer;
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;
      pragma Debug (ASSERT (TE_Ac.all in Timed_Handler_Timed_Event'Class));

      Kernel.Enter_Critic_Section (Flags);

      if TE_Ac = null or else TE_Ac.Magic /= INITIALIZED then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      if not TE_Ac.Enabled then

         TE_Ac.Enabled := True;

         if TE_Ac.Active and then not TE_Ac.Queued then
            --  Out of the event queue: that means the expiration time
            --  was reached while the event was disabled. The handler
            --  has to be executed now.
            Internals.Execute_Timed_Handler (TE_Ac,
                                             From_Timer_Int_Handler => False,
                                             Now => HWI.Get_HWTime);
         end if;

      end if;  --  not TE_Ac.Enabled

      Kernel.Leave_Critic_Section (Flags);

      return 0;
   end MaRTE_Timed_Handler_Enable;

   -----------------------------------------
   --  MaRTE_Timed_Handler_Global_Enable  --
   -----------------------------------------
   function MaRTE_Timed_Handler_Global_Enable return Int is
      Flags : Integer;
      TE_Ac : Timed_Handler_TE_Ac;
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);

      Internals.Execution_Allowed := True;

      --  Execute every pending handler
      TE_Ac := Timed_Handler_TE_Ac
        (Kernel.Timed_Events_Lists.Head (Internals.Pending_Handlers_List));
      while (TE_Ac /= null) loop
         pragma Debug (ASSERT (TE_Ac.Pending and not TE_Ac.Queued));
         TE_Ac.Pending := False;
         Internals.Execute_Timed_Handler (TE_Ac,
                                          From_Timer_Int_Handler => False,
                                          Now => HWI.Get_HWTime);
         TE_Ac := Timed_Handler_TE_Ac
           (Kernel.Timed_Events_Lists.Next (Kernel.Timed_Event_Ac (TE_Ac)));
      end loop;

      Kernel.Leave_Critic_Section (Flags);

      return 0;
   end MaRTE_Timed_Handler_Global_Enable;

   ------------------------------------------
   --  MaRTE_Timed_Handler_Global_Disable  --
   ------------------------------------------
   function MaRTE_Timed_Handler_Global_Disable return Int is
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;

      Internals.Execution_Allowed := False;

      return 0;
   end MaRTE_Timed_Handler_Global_Disable;

   ---------------------------------------
   --  MaRTE_Timed_Handler_Get_Overrun  --
   ---------------------------------------
   function MaRTE_Timed_Handler_Get_Overrun (TE_Ac : in Timed_Handler_TE_Ac)
                                            return Int is
   begin
      return Int (TE_Ac.Overrun_Count);
   end MaRTE_Timed_Handler_Get_Overrun;

   -----------
   --  Set  --
   -----------
   procedure Set (TE  : in out Timed_Handler_Timed_Event;
                  Now : in     HWI.HWTime) is
      use type HWI.HWTime, Kernel.Task_Id;
   begin
      if not Configuration_Parameters.Use_Timed_Handlers'First then
         return;
      end if;
      pragma Debug (ASSERT (not HWI.Are_Interrupts_Enabled));
      pragma Debug (ASSERT (TE in Timed_Handler_Timed_Event'Class));
      pragma Debug (ASSERT (TE.Magic = INITIALIZED));
      pragma Debug (ASSERT (TMRI.Clock_Id_OK (TE.Base_Clock)));

      --  Remove "old" event if active
      if TE.Queued then
         --  Queued event waiting for expiration time: remove from
         --  timed events queue
         if (TE.Base_Clock = Kernel.Timers.CLOCK_REALTIME or
               TE.Base_Clock = Kernel.Timers.CLOCK_MONOTONIC) then
            --  Standard event
            SCHD.Remove_Timed_Event (TE'Unrestricted_Access);
         else
            --  timed handler based on CPU time clock
            CPU_TE_LST.Dequeue
              (TE'Unrestricted_Access,
               TMRI.To_Task_Id (TE.Base_Clock).CPU_Time_TEs_Q);
         end if;
      elsif TE.Active then
         --  Expired event whose handler hasn't been executed yet,
         --  because either the event is disabled or handler execution
         --  is disallowed. This situation is detected because the
         --  event is active but is not queued.
         if TE.Pending then
            --  Remove from the 'Pending_Handlers_List'
            pragma Debug
              (ASSERT (Kernel.Timed_Events_Lists.Is_In_The_List
                       (TE'Unrestricted_Access,
                        Internals.Pending_Handlers_List)));
            Kernel.Timed_Events_Lists.Dequeue
              (TE'Unrestricted_Access, Internals.Pending_Handlers_List);
         end if;
         --  For an expired and disabled event nothing needs to be
         --  done, it will be mark as "queued" at the end of this
         --  procedure.
      end if;

      --  Set "new" timed event expiration time
      if (TE.Base_Clock = Kernel.Timers.CLOCK_REALTIME or
            TE.Base_Clock = Kernel.Timers.CLOCK_MONOTONIC) then
         --  timed handler based on standard clock
         TMRI.Set_Time_Event
           (TE_Ac     => TE'Unrestricted_Access,
            T         => TE.T,
            Clock     => TE.Base_Clock,
            Abs_Time  => (TE.Options and ABSOLUTE_TIMER) /= 0 and then
                         (TE.Options and PERIODIC_HANDLER) = 0,
            Now       => Now);

      else
         pragma Debug (ASSERT ((TE.Options and PERIODIC_HANDLER) = 0));
         --  timed handler based on CPU clock
         TMRI.Set_CPU_Time_Event
           (TE_Ac     => TE'Unrestricted_Access,
            T         => TE.T,
            CPU_Clock => TE.Base_Clock,
            Abs_Time  => (TE.Options and ABSOLUTE_TIMER) /= 0 and then
                         (TE.Options and PERIODIC_HANDLER) = 0,
            Now       => Now);
      end if;

      --  Program event or execute handler
      TE.Enabled := True;
      TE.Active  := True;
      TE.Queued  := False;
      TE.Pending := False;

      if TE.T < Now then
         --  Expiration time is in the past: Execute handler
         Internals.Execute_Timed_Handler (TE'Unrestricted_Access,
                                          From_Timer_Int_Handler => False,
                                          Now => Now);
      else
         --  Program event
         TE.Queued := True;
         if (TE.Base_Clock = Kernel.Timers.CLOCK_REALTIME or
               TE.Base_Clock = Kernel.Timers.CLOCK_MONOTONIC) then
            --  Enqueue standard event
            SCHD.Enqueue_Standard_Event_And_Update_Timer
              (TE'Unrestricted_Access, Now);
         elsif TMRI.To_Task_Id (TE.Base_Clock) = SCHD.Self then
            --  The event's base clock is the CPU clock of the running task
            CPU_TE_LST_ORDER.Enqueue_In_Order
              (TE'Unrestricted_Access, SCHD.Self.CPU_Time_TEs_Q);
            TE_T.Update_Timer_Due_To_New_CPU_Event (TE'Unrestricted_Access,
                                                       Now);
         else
            --  The timer's base clock is the CPU clock of other task
            CPU_TE_LST_ORDER.Enqueue_In_Order
              (TE'Unrestricted_Access,
               TMRI.To_Task_Id (TE.Base_Clock).CPU_Time_TEs_Q);
         end if;
      end if;
   end Set;

end Kernel.Timed_Handlers.Operations;
