------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--              'K e r n e l . T i m e r s . I n t e r n a l s'
--
--                                 Body
--
--
--  File 'k-timers-internals.adb'                                      By MAR.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
-------------------------------------------------------------------------------
with Ada.Unchecked_Conversion;

with Kernel.Signals.Internals;
with Kernel.Scheduler;
with Kernel.Timed_Events_And_Timer;
with Configuration_Parameters;

package body Kernel.Timers.Internals is

   package SCHD  renames Kernel.Scheduler;
   package TE_T  renames Kernel.Timed_Events_And_Timer;
   package SI    renames Kernel.Signals.Internals;

   use Hardware_Interface;
   use Kernel.Tasks_Lists;
   use Kernel.CPU_Time_Timed_Events_Lists;
   use Kernel.CPU_Time_Timed_Events_Lists_Order;
   use Option_Sets;

   ----------------------------
   --  Is_CPU_Clock_Of_Self  --
   ----------------------------
   function Is_CPU_Clock_Of_Self (Clock : Clock_Id) return Boolean is
   begin
      return (Clock = CLOCK_THREAD_CPUTIME_ID or else
              SCHD.Self = To_Task_Id (Clock));
   end Is_CPU_Clock_Of_Self;

   -----------------
   -- Clock_Id_OK --
   -----------------
   function Clock_Id_OK (Clock : Clock_Id) return Boolean is
   begin
      return (Clock = CLOCK_REALTIME   or else
              Clock = CLOCK_MONOTONIC  or else
              (Configuration_Parameters.Use_CPU_Time_Clocks_And_Timers and then
               (Clock = CLOCK_THREAD_CPUTIME_ID or else
                Task_OK (To_Task_Id (Clock), NEW_APPSCHED))));
   end Clock_Id_OK;
   pragma Inline (Clock_Id_OK);

   -----------------
   -- Timer_Id_OK --
   -----------------
   function Timer_Id_OK (Timer : Timer_Id) return Boolean is
   begin
      return Timer /= null and then Timer.Magic = INITIALIZED;
   end Timer_Id_OK;
   pragma Inline (Timer_Id_OK);

   ---------------------
   --  Timer_Expired  --
   ---------------------
   --
   --  Called from the timer interrupt handler
   --  'SCHD.Timer_Interrupt_Handler'.
   procedure Timer_Expired (Timer : in Timer_Id; At_Time : in HWI.HWTime) is
   begin
      --  Is it a one-shot timer ?
      if Timer.State.Interval = 0 then
         --  It is a one-shot timer
         Timer.Armed := False;

      else  --  It is a periodic timer... then rearms it.

         if (Timer.Base_Clock = CLOCK_REALTIME or
             Timer.Base_Clock = CLOCK_MONOTONIC) then
            --  Timer NOT based on a execution-time clock
            Timer.T := Timer.T + Timer.State.Interval;
            TE_T.Enqueue_Standard_Event_Without_Updating_Timer
              (Timed_Event_Ac (Timer));

         else
            --  Execution-Time timer.
            pragma Assert (To_Task_Id (Timer.Base_Clock) = SCHD.Self);
            Timer.CPU_Time := Timer.CPU_Time + Timer.State.Interval;
            Timer.T := Timer.T + Timer.State.Interval;
            Enqueue_In_Order (CPU_Time_Timed_Event_Ac (Timer),
                              SCHD.Self.CPU_Time_TEs_Q);
         end if;
      end if;

      --  Perform timer action
      case Timer.Event.Event_Notification is
         when Kernel.Signals.SIGNAL_NOTIFICATION =>
            if (not
                SI.Is_Reserved_Signal_Instance_Pending (Timer.Siginst_Ac)) then
               Timer.Overruns := 0;
               SI.Send_Signal_To_Process (Timer.Siginst_Ac);
            else
               Timer.Overruns := Timer.Overruns + 1;
            end if;
         when Kernel.Signals.NO_NOTIFICATION =>
            null;
      end case;
   end Timer_Expired;

   --------------------
   --  Disarm_Timer  --
   --------------------
   procedure Disarm_Timer (Timer : in Timer_Id) is
   begin
      pragma Assert (Clock_Id_OK (Timer.Base_Clock));
      Timer.Armed := False;
      if (Timer.Base_Clock = CLOCK_REALTIME or
          Timer.Base_Clock = CLOCK_MONOTONIC) then
         TE_T.Remove_Standard_Event (Timed_Event_Ac (Timer));

      else
         --  Execution-Time timer.
         Dequeue (CPU_Time_Timed_Event_Ac (Timer),
                  To_Task_Id (Timer.Base_Clock).CPU_Time_TEs_Q);
      end if;
   end Disarm_Timer;

   -----------------
   --  Arm_Timer  --
   -----------------
   procedure Arm_Timer (Timer     : in  Timer_Id;
                        Options   : in  Timer_Options;
                        New_State : in  Timer_State;
                        Now       : in  HWI.HWTime) is
   begin
      pragma Assert (Clock_Id_OK (Timer.Base_Clock));
      --  If armed disarm first
      if Timer.Armed then
         Disarm_Timer (Timer);
      end if;
      --  Initialize timer fields
      Timer.State := New_State;
      Timer.Armed := True;

      if (Timer.Base_Clock = CLOCK_REALTIME or
          Timer.Base_Clock = CLOCK_MONOTONIC) then
         Timer.T := Timer.State.Initial;
         if ABSOLUTE_TIMER <= Options then
            --  Absolute timer
            if Timer.Base_Clock = CLOCK_REALTIME then
               if Timer.T < Realtime_Clock_Offset then
                  Timer.T := 0;
               else
                  Timer.T := Timer.T - Realtime_Clock_Offset;
               end if;
            end if;
         else
            --  Relative timer
            Timer.T := Now + Timer.State.Initial;
         end if;
         --  At this point in 'Timer.T' is the timer expiration ABSOLUTE
         --  time respect the hardware clock (monotonic clock that measures
         --  time from system reset).
         if Timer.T < Now then
            --  The expiration time is before the current time, then it is
            --  normalized to expire just now.
            Timer.T := Now;
         end if;
         TE_T.Enqueue_Standard_Event_And_Update_Timer (Timed_Event_Ac (Timer),
                                                       Now);

      elsif To_Task_Id (Timer.Base_Clock) = SCHD.Self then
         --  The timer's base clock is the CPU clock of the running task
         if ABSOLUTE_TIMER <= Options then
            Timer.CPU_Time := Timer.State.Initial;
            if Timer.State.Initial < SCHD.Self.Used_CPU_Time then
               Timer.T := SCHD.Self.Running_Starting_Time
                 - (SCHD.Self.Used_CPU_Time - Timer.State.Initial);
            else
               Timer.T := SCHD.Self.Running_Starting_Time
                 + Timer.State.Initial - SCHD.Self.Used_CPU_Time;
            end if;
         else
            Timer.CPU_Time := SCHD.Self.Used_CPU_Time +
              Now - SCHD.Self.Running_Starting_Time + Timer.State.Initial;
            Timer.T := Now + Timer.State.Initial;
         end if;
         Enqueue_In_Order (CPU_Time_Timed_Event_Ac (Timer),
                           SCHD.Self.CPU_Time_TEs_Q);
         TE_T.Update_Timer_Due_To_New_CPU_Event (Timed_Event_Ac (Timer), Now);

      else
         --  The timer's base clock is the CPU clock of other task
         if ABSOLUTE_TIMER <= Options then
            Timer.CPU_Time := Timer.State.Initial;
         else
            Timer.CPU_Time := To_Task_Id (Timer.Base_Clock).Used_CPU_Time
              + Timer.State.Initial;
         end if;
         Enqueue_In_Order (CPU_Time_Timed_Event_Ac (Timer),
                           To_Task_Id (Timer.Base_Clock).CPU_Time_TEs_Q);
      end if;
   end Arm_Timer;

   -----------------------
   --  Get_Timer_State  --
   -----------------------
   function Get_Timer_State (Timer : Timer_Id;
                             Now : HWI.HWTime) return Timer_State is
      TS : Timer_State;
   begin
      pragma Assert (Clock_Id_OK (Timer.Base_Clock));
      if Timer.Armed then
         TS.Interval := Timer.State.Interval;
         if (Timer.Base_Clock = CLOCK_REALTIME or
             Timer.Base_Clock = CLOCK_MONOTONIC) then
            if Timer.T < Now then
               TS.Initial := 0;
            else
               TS.Initial := Timer.T - Now;
            end if;

         elsif To_Task_Id (Timer.Base_Clock) = SCHD.Self then
            --  The timer's base clock is the CPU clock of the running task
            if Timer.T - SCHD.Self.Used_CPU_Time <
              (Now - SCHD.Self.Running_Starting_Time) then
               TS.Initial := 0;
            else
               TS.Initial := Timer.T - SCHD.Self.Used_CPU_Time -
                 (Now - SCHD.Self.Running_Starting_Time);
            end if;

         else
            --  The timer's base clock is the CPU clock of other task
            if (Timer.CPU_Time <
                To_Task_Id (Timer.Base_Clock).Used_CPU_Time) then
               TS.Initial := 0;
            else
               TS.Initial := Timer.CPU_Time
                 - To_Task_Id (Timer.Base_Clock).Used_CPU_Time;
            end if;
         end if;

      else
         --  timer disarmed
         TS.Initial := 0;
         TS.Interval := 0;
      end if;

      return TS;
   end Get_Timer_State;

   ----------------------
   --  Set_Time_Event  --
   ----------------------
   --  Sets 'TE_Ac.T' as an ABSOLUTE time respect the hardware clock (monotonic
   --  clock that measures time from system reset).
   --  'TE_Ac.T' is NOT normalized: can return a value of time before the
   --  current time
   procedure Set_Time_Event (TE_Ac    : in Kernel.Timed_Event_Ac;
                             T        : in HWI.HWTime;
                             Abs_Time : in Boolean;
                             Clock    : in Clock_Id;
                             Now      : in HWI.HWTime) is
   begin
      pragma Assert (Clock_Id_OK (Clock));
      pragma Assert (Clock = CLOCK_REALTIME or Clock = CLOCK_MONOTONIC);

      if Abs_Time then
         --  Absolute time
         if Clock = CLOCK_REALTIME then
            if T < Realtime_Clock_Offset then
                  TE_Ac.T := 0;
            else
               TE_Ac.T := T - Realtime_Clock_Offset;
            end if;
         else  --  monotonic clock
            TE_Ac.T := T;
         end if;
      else
         --  Relative time
         TE_Ac.T := Now + T;
      end if;
   end Set_Time_Event;

   --------------------------
   --  Set_CPU_Time_Event  --
   --------------------------
   procedure Set_CPU_Time_Event (TE_Ac     : in Kernel.CPU_Time_Timed_Event_Ac;
                                 T         : in HWI.HWTime;
                                 CPU_Clock : in Clock_Id;
                                 Abs_Time  : in Boolean;
                                 Now       : in HWI.HWTime) is
   begin
      pragma Assert (Clock_Id_OK (CPU_Clock));
      pragma Assert (CPU_Clock /= CLOCK_REALTIME and
                       CPU_Clock /= CLOCK_MONOTONIC);

      if To_Task_Id (CPU_Clock) = SCHD.Self then

         --  CPU clock of the running task
         if Abs_Time then
            --  Absolute time
            TE_Ac.CPU_Time := T;
            if T < SCHD.Self.Used_CPU_Time then
               TE_Ac.T := SCHD.Self.Running_Starting_Time
                 - (SCHD.Self.Used_CPU_Time - T);
            else
               TE_Ac.T := SCHD.Self.Running_Starting_Time
                 + T - SCHD.Self.Used_CPU_Time;
            end if;
         else
            --  Relative time
            TE_Ac.CPU_Time := SCHD.Self.Used_CPU_Time +
              Now - SCHD.Self.Running_Starting_Time + T;
            TE_Ac.T := Now + T;
         end if;

      else
         --  CPU clock of another task
         --  TE_Ac.T si also set since some functions check if TE_Ac.T < now to
         --  perform the action inmediately in that case
         TE_Ac.T := HWI.HWTime'Last;
         if Abs_Time then
            TE_Ac.CPU_Time := T;
            if T < To_Task_Id (CPU_Clock).Used_CPU_Time then
               TE_Ac.T := HWI.HWTime'First;  --  time has passed
            end if;
         else
            TE_Ac.CPU_Time := To_Task_Id (CPU_Clock).Used_CPU_Time + T;
         end if;
      end if;
   end Set_CPU_Time_Event;

   ----------------------------------------
   --  Time_Remaining_To_CPU_Time_Event  --
   ----------------------------------------
   --  Returns the execution time interval that remains until the CPU even
   --  TE_Ac would expire
   function Time_Remaining_To_CPU_Time_Event
     (TE_Ac     : Kernel.CPU_Time_Timed_Event_Ac;
      CPU_Clock : Clock_Id;
      Now       : HWI.HWTime)
      return HWI.HWTime is
   begin
      pragma Assert (Clock_Id_OK (CPU_Clock));
      pragma Assert (CPU_Clock /= CLOCK_REALTIME and
                       CPU_Clock /= CLOCK_MONOTONIC);

      if To_Task_Id (CPU_Clock) = SCHD.Self then

         --  CPU clock of the running task
         if TE_Ac.T - SCHD.Self.Used_CPU_Time <
           (Now - SCHD.Self.Running_Starting_Time) then
            return 0;
         else
            return TE_Ac.T - SCHD.Self.Used_CPU_Time -
              (Now - SCHD.Self.Running_Starting_Time);
         end if;

      else
         --  The timer's base clock is the CPU clock of other task
         if (TE_Ac.CPU_Time <
               To_Task_Id (CPU_Clock).Used_CPU_Time) then
            return 0;
         else
            return TE_Ac.CPU_Time - To_Task_Id (CPU_Clock).Used_CPU_Time;
         end if;
      end if;
   end Time_Remaining_To_CPU_Time_Event;

end Kernel.Timers.Internals;
