------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--                        'K e r n e l . T i m e r s'
--
--                                   Body
--
--
--
--  File 'k-timers.adb'                                                By MAR.
--
--  Timers management (including CPU-Time timers).
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with Ada.Unchecked_Conversion;
with Kernel.Timers.Internals;
with Kernel.Timer_Timed_Events_Pool;
with Kernel.Scheduler;
with Kernel.Tasks_Operations.Internals;
with Kernel.Signals.Internals;
with Configuration_Parameters;
with Hardware_Interface;
pragma Elaborate_All (Hardware_Interface);

package body Kernel.Timers is

   use type HWI.HWTime;
   package TOI  renames Kernel.Tasks_Operations.Internals;
   package SCHD renames Kernel.Scheduler;
   package SI   renames Kernel.Signals.Internals;

   use Kernel.Tasks_Lists;
   use Kernel.CPU_Time_Timed_Events_Lists;
   use Kernel.CPU_Time_Timed_Events_Lists_Order;
   use Kernel.Signals.Global.Signal_Instances_Lists;

   use type Int;

   --------------------------------------------------
   -- Conversions between 'Task_Id' and 'Clock_Id' --
   --------------------------------------------------
   function To_Task_Id is
     new Ada.Unchecked_Conversion (Clock_Id, Kernel.Task_Id);
   function To_Clock_Id is
     new Ada.Unchecked_Conversion (Kernel.Task_Id, Clock_Id);

   ---------------------------------------------
   -- Conversion between 'Timer_Id' and 'Int' --
   ---------------------------------------------
   function Timer_Id_To_Int is new Ada.Unchecked_Conversion (Timer_Id, Int);

   ----------------
   --  Set_Time  --
   ----------------
   procedure Set_Time (Clock : in Clock_Id;
                       Value : in HWI.HWTime) is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if (Clock = CLOCK_MONOTONIC or else
          not Internals.Clock_Id_OK (Clock)) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return;
      end if;
      TOI.Reset_POSIX_Error;

      if Clock = CLOCK_REALTIME then
         Realtime_Clock_Offset := Value - HWI.Get_HWTime;

      elsif (Clock = CLOCK_THREAD_CPUTIME_ID or else
             To_Task_Id (Clock) = SCHD.Self) then
         --  The clock is associated with the running task
         SCHD.Self.Running_Starting_Time := HWI.Get_HWTime;
         SCHD.Self.Used_CPU_Time := Value;

      else
         --  The clock is associated with a task other than the running
         To_Task_Id (Clock).Used_CPU_Time := Value;
      end if;
      Kernel.Leave_Critic_Section (Flags);
   end Set_Time;

   ----------------
   --  Get_Time  --
   ----------------
   function Get_Time (Clock : Clock_Id) return HWI.HWTime is
      Flags    : Integer;
      Ret_Time : HWI.HWTime;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Internals.Clock_Id_OK (Clock) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return 0;
      end if;
      TOI.Reset_POSIX_Error;

      if Clock = CLOCK_REALTIME then
         Ret_Time := HWI.Get_HWTime + Realtime_Clock_Offset;

      elsif Clock = CLOCK_MONOTONIC then
         Ret_Time := HWI.Get_HWTime;

      elsif (Clock = CLOCK_THREAD_CPUTIME_ID or else
             To_Task_Id (Clock) = SCHD.Self) then
         Ret_Time := (SCHD.Self.Used_CPU_Time + HWI.Get_HWTime)
           - SCHD.Self.Running_Starting_Time;

      else --  The clock is associated with a task other than the running
         Ret_Time := To_Task_Id (Clock).Used_CPU_Time;

      end if;
      Kernel.Leave_Critic_Section (Flags);
      return Ret_Time;
   end Get_Time;

   ---------------------
   --  Get_Time_Gnat  --
   ---------------------
   --
   --  Used by GNARL (s-osprim.adb)
   function Get_Time_Gnat return Duration is
      Flags    : Integer;
      Ret_Time : Duration;
   begin
      Kernel.Enter_Critic_Section (Flags);
      TOI.Reset_POSIX_Error;

      Ret_Time :=
        HWI.HWTime_To_Duration (HWI.Get_HWTime + Realtime_Clock_Offset);

      Kernel.Leave_Critic_Section (Flags);
      return Ret_Time;
   end Get_Time_Gnat;

   ----------------------
   --  Get_Resolution  --
   ----------------------
   function Get_Resolution (Clock : Clock_Id) return Duration is
      Flags    : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Internals.Clock_Id_OK (Clock) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return 0.0;
      end if;
      TOI.Reset_POSIX_Error;

      Kernel.Leave_Critic_Section (Flags);

      return 1.0 / Duration (HWI.Get_HWClock_Frequency);
   end Get_Resolution;

   ---------------------------
   -- Pthread_Getcpuclockid --
   ---------------------------
   function Pthread_Getcpuclockid (T     : in     Task_Id;
                                   Clock : access Clock_Id) return Int is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Configuration_Parameters.Use_CPU_Time_Clocks_And_Timers then
         Kernel.Leave_Critic_Section (Flags);
         return OPERATION_NOT_SUPPORTED;
      end if;
      if not Task_OK (T, NEW_APPSCHED) then
         Kernel.Leave_Critic_Section (Flags);
         return NO_SUCH_PROCESS;
      end if;

      Clock.all := To_Clock_Id (T);

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Pthread_Getcpuclockid;

   --------------------
   --  Create_Timer  --
   --------------------
   function Create_Timer (Clock         : Clock_Id;
                          Event         : Signals.Signal_Event;
                          Default_Event : Boolean := False) return Timer_Id is
      Timer : Timer_Id;
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Internals.Clock_Id_OK (Clock) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return null;
      end if;

      Timer := Kernel.Timer_Timed_Events_Pool.Request_Timer_Timed_Event;
      if Timer = null then
         TOI.Set_POSIX_Error (RESOURCE_TEMPORARILY_UNAVAILABLE);
         Kernel.Leave_Critic_Section (Flags);
         return null;
      end if;
      Timer.Siginst_Ac := SI.Request_And_Reserve_Signal_Instance;
      if Timer.Siginst_Ac = null then
         TOI.Set_POSIX_Error (RESOURCE_TEMPORARILY_UNAVAILABLE);
         Kernel.Leave_Critic_Section (Flags);
         return null;
      end if;
      TOI.Reset_POSIX_Error;

      if Clock = CLOCK_THREAD_CPUTIME_ID then
         Timer.Base_Clock      := To_Clock_Id (SCHD.Self);
      else
         Timer.Base_Clock      := Clock;
      end if;
      Timer.Armed           := False;
      Timer.Event           := Event;
      if Signals."=" (Timer.Event.Event_Notification,
                      Signals.SIGNAL_NOTIFICATION) then
         Timer.Siginst_Ac.Signo   := Timer.Event.Event_Signal;
         Timer.Siginst_Ac.Code    := Signals.SI_TIMER;
         if Default_Event then
            Timer.Siginst_Ac.Value := Timer_Id_To_Int (Timer);
         else
            Timer.Siginst_Ac.Value := Timer.Event.Event_Sigval;
         end if;
      end if;
      Timer.Overruns        := 0;
      Timer.Magic           := INITIALIZED;

      Kernel.Leave_Critic_Section (Flags);
      return Timer;
   end Create_Timer;

   --------------------
   --  Timer_Delete  --
   --------------------
   function Timer_Delete (Timer : in Timer_Id) return Int is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Internals.Timer_Id_OK (Timer) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return -1;
      end if;
      TOI.Reset_POSIX_Error;

      if Timer.Armed and then Internals.Clock_Id_OK (Timer.Base_Clock) then
         Internals.Disarm_Timer (Timer);
      end if;
      Timer.Magic := FINALIZED;
      Kernel.Timer_Timed_Events_Pool.Set_Timer_Timed_Event_Free (Timer);
      SI.Cancel_Reserve_And_Release_Signal_Instance (Timer.Siginst_Ac);

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Timer_Delete;

   -----------------
   --  Arm_Timer  --
   -----------------
   procedure Arm_Timer (Timer     : in  Timer_Id;
                        Options   : in  Timer_Options;
                        New_State : in  Timer_State;
                        Old_State : out Timer_State) is
      Flags : Integer;
      Now   : HWI.HWTime;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if not Internals.Timer_Id_OK (Timer) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return;
      end if;
      pragma Assert (Timer.Base_Clock /= CLOCK_THREAD_CPUTIME_ID);
      TOI.Reset_POSIX_Error;

      Now := HWI.Get_HWTime;
      Old_State := Internals.Get_Timer_State (Timer, Now);
      Internals.Arm_Timer (Timer, Options, New_State, Now);

      Kernel.Leave_Critic_Section (Flags);
   end Arm_Timer;

   -----------------
   --  Arm_Timer  --
   -----------------
   procedure Arm_Timer (Timer     : in Timer_Id;
                        Options   : in Timer_Options;
                        New_State : in Timer_State) is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if not Internals.Timer_Id_OK (Timer) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return;
      end if;
      pragma Assert (Timer.Base_Clock /= CLOCK_THREAD_CPUTIME_ID);
      TOI.Reset_POSIX_Error;

      Internals.Arm_Timer (Timer, Options, New_State, HWI.Get_HWTime);

      Kernel.Leave_Critic_Section (Flags);
   end Arm_Timer;

   -----------------------
   --  Get_Timer_State  --
   -----------------------
   function Get_Timer_State (Timer : Timer_Id) return Timer_State is
      TS : Timer_State;
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Internals.Timer_Id_OK (Timer) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return Timer_State'(0, 0);
      end if;
      TOI.Reset_POSIX_Error;

      TS := Internals.Get_Timer_State (Timer, HWI.Get_HWTime);

      Kernel.Leave_Critic_Section (Flags);
      return TS;
   end Get_Timer_State;

   --------------------
   --  Disarm_Timer  --
   --------------------
   procedure Disarm_Timer (Timer : in Timer_Id) is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);

      if not Internals.Timer_Id_OK (Timer) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return;
      end if;
      TOI.Reset_POSIX_Error;

      if Timer.Armed then
         Internals.Disarm_Timer (Timer);
      end if;

      Kernel.Leave_Critic_Section (Flags);
   end Disarm_Timer;

   ------------------------
   --  Timer_Getoverrun  --
   ------------------------
   function Timer_Getoverrun (Timer : Timer_Id) return Int is
      Flags    : Integer;
      Overruns : Natural;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if not Internals.Timer_Id_OK (Timer) then
         TOI.Set_POSIX_Error (INVALID_ARGUMENT);
         Kernel.Leave_Critic_Section (Flags);
         return -1;
      end if;
      TOI.Reset_POSIX_Error;

      Overruns := Timer.Overruns;
      Kernel.Leave_Critic_Section (Flags);

      return Int (Overruns);
   end Timer_Getoverrun;

begin
   if HWI.RTC_Available then
      Realtime_Clock_Offset := HWI.RTC_HWTime_Since_Epoch;
      Realtime_Clock_Offset := Realtime_Clock_Offset - HWI.Get_HWTime;
   else
      --  CLOCK_REALTIME and CLOCK_MONOTONIC will start being the same.
      Realtime_Clock_Offset := 0;
   end if;
end Kernel.Timers;
