------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--    'A d a . A p p l i c a t i o n _ S c h e d u l i n g . S u p p o r t'
--
--                                  Body
--
--
--  File 'a-apscsu.adb'                                                By MAR
--                                                                    and MGH
--
--
--  Suport for runtime and compiler implementation of
--  Application-Defined Scheduling.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with System.OS_Interface; --  To use 'Self_Key'
with System.Task_Primitives.Operations;
with System.Tasking;
with Ada.Unchecked_Conversion;
with Ada.Exceptions;
with Ada.Real_Time.Delays; -- To use 'To_Duration (T : Real_Time.Time)'

with Basic_Integer_Types; use Basic_Integer_Types;
with Hardware_Interface;
with Kernel.Signals.POSIX_Functions;
with Kernel.Application_Scheduling_Data;
with Kernel.Tasks_Operations;
with Kernel.Tasks_Operations.Application_Scheduler;
with Kernel.Scheduler;
with Kernel.Signals.POSIX_Functions;

with Kernel_Console;

package body Ada.Application_Scheduling.Support is

   package APPSCHD_DATA renames Kernel.Application_Scheduling_Data;
   package TO           renames Kernel.Tasks_Operations;
   package TO_APPSCHD   renames Kernel.Tasks_Operations.Application_Scheduler;
   package SCHD         renames Kernel.Scheduler;
   package SGNL_POSIX   renames Kernel.Signals.POSIX_Functions;
   package HWI          renames Hardware_Interface;
   use type Kernel.Task_Id;


   ----------------------------------------------
   -- Conversions between 'Kernel.Task_Id' and --
   -- 'Ada.Task_Identification.Task_Id'        --
   ----------------------------------------------
   --
   function To_MaRTE_Task_Id (Ada_Task_Id : Ada.Task_Identification.Task_Id)
                             return Kernel.Task_Id;
   function To_MaRTE_Task_Id (Ada_Task_Id : Ada.Task_Identification.Task_Id)
                             return Kernel.Task_Id is

      function Pthread_T_To_MaRTE_Task_Id is new Ada.Unchecked_Conversion
        (System.OS_Interface.pthread_t, Kernel.Task_Id);
      function Convert_Ids is new Ada.Unchecked_Conversion
        (Ada.Task_Identification.Task_Id, System.Tasking.Task_ID);
   begin
      return
        Pthread_T_To_MaRTE_Task_Id
        (System.Task_Primitives.Operations.Get_Thread_Id
         (Convert_Ids (Ada_Task_Id)));
   end To_MaRTE_Task_Id;

   function To_Ada_Task_Id (Kernel_Task_Id : Kernel.Task_Id)
                           return Ada.Task_Identification.Task_Id;
   function To_Ada_Task_Id (Kernel_Task_Id : Kernel.Task_Id)
                           return Ada.Task_Identification.Task_Id is
      function Address_To_Ada_Task_Id is
        new Unchecked_Conversion (System.Address,
                                  Ada.Task_Identification.Task_Id);
   begin
      return Address_To_Ada_Task_Id
        (Kernel_Task_Id.Specific_Data (System.OS_Interface.Self_Key));
   end To_Ada_Task_Id;

   ---------------
   --  Check_NZ --
   ---------------
   procedure Check_NZ (Result : Int);
   procedure Check_NZ (Result : Int) is
   begin
      if Result /= 0 then
         raise MaRTE_AppSched_Error;
      end if;
   end Check_NZ;

   ----------------------------------------------------------------------------
   ----  Generic Scheduler Task Type  -----------------------------------------
   ----------------------------------------------------------------------------
   task type Application_Scheduler_Task_Type
     (Prio             : System.Any_Priority;
      Sched            : access Scheduler'Class;
      AppSched_Data_Ac : Kernel.AppScheduler_Data_Base_Ac) is
      pragma Priority (Prio);
   end Application_Scheduler_Task_Type;

   type Application_Scheduler_Task_Type_Ac is
     access Application_Scheduler_Task_Type;


   ----------------------------------------------------------------------------
   ----  Manage Schedulers and Appscheduled Tasks  ----------------------------
   ----------------------------------------------------------------------------

   Scheduler_Signo : Kernel.Signals.Signal := Kernel.Signals.Signal'Last;
   --  Used to assign diferent signal numbers with schedulers.

   type Scheduler_Associaton is record
      Sched_Task_Id : Kernel.Task_Id;
      Signo         : Kernel.Signals.Signal;
      Installed     : Boolean;
   end record;
   Associated_Schedulers : array (System.Any_Priority) of Scheduler_Associaton
     := (others => (Sched_Task_Id => null,
                    Signo         => Kernel.Signals.Signal'Last,
                    Installed     => False));

   -------------------------
   --  Install_Scheduler  --
   -------------------------
   procedure Install_Scheduler (Sched : access Scheduler'Class;
                                Prio  : in     System.Any_Priority) is
      AppSched_Data_Ac : Kernel.AppScheduler_Data_Base_Ac;
      use type Kernel.AppScheduler_Data_Base_Ac;
      T : Application_Scheduler_Task_Type_Ac;
      use type Kernel.Signals.Signal;
   begin
      --  Get an 'AppScheduler_Data' object (necessary to create an
      --  application-scheduler task)
      AppSched_Data_Ac := Kernel.AppScheduler_Data_Base_Ac
        (APPSCHD_DATA.Request_AppScheduler_Data);
      if AppSched_Data_Ac = null then
         --  Impossible create more application scheduler tasks
         raise Reached_Maximum_Number_Of_Application_Schedulers;
      end if;

      --  Create scheduler task
      T := new Application_Scheduler_Task_Type (Prio,
                                                Sched,
                                                AppSched_Data_Ac);

      --  Associate with priority level
      Associated_Schedulers (Prio) :=
        (Sched_Task_Id => To_MaRTE_Task_Id (T'Identity),
         Signo         => Scheduler_Signo,
         Installed     => True);

      Scheduler_Signo := Scheduler_Signo - 1;
   end Install_Scheduler;

   ---------------------------
   --  Scheduler_Installed  --
   ---------------------------
   function Scheduler_Installed (Prio : in System.Any_Priority)
                                return Boolean is
   begin
      return Associated_Schedulers (Prio).Installed;
   end Scheduler_Installed;

   --------------------------------
   --  Attach_Task_To_Scheduler  --
   --------------------------------
   procedure Attach_Task_To_Scheduler
     (Tid  : in Ada.Task_Identification.Task_Id;
      Prio : in System.Any_Priority) is
      Sched_Param : aliased TO.Sched_Param;
      Policy      : aliased TO.Scheduling_Policies;
   begin
      if not Associated_Schedulers (Prio).Installed then
         raise Not_Scheduler_At_Priority;
      end if;

      --  Set appscheduler attribute
      Check_NZ (TO_APPSCHD.Pthread_Setappscheduler
                                 (To_MaRTE_Task_Id (Tid),
                                  Associated_Schedulers (Prio).Sched_Task_Id));

      --  Get current scheduling parameters
      Check_NZ (TO.Pthread_Getschedparam (To_MaRTE_Task_Id (Tid),
                                          Policy'Access,
                                          Sched_Param'Access));

      --  Set new policy
      Check_NZ (TO.Pthread_Setschedparam (To_MaRTE_Task_Id (Tid),
                                          Kernel.SCHED_APP,
                                          Sched_Param'Access));
   end Attach_Task_To_Scheduler;


   ----------------------------------
   --  Detach_Task_From_Scheduler  --
   ----------------------------------
   procedure Detach_Task_From_Scheduler
     (Tid  : in Ada.Task_Identification.Task_Id;
      Prio : in System.Any_Priority) is
      Sched_Param : aliased TO.Sched_Param;
      Policy      : aliased TO.Scheduling_Policies;
      Sched       : aliased Kernel.Task_Id;
   begin
      if not Associated_Schedulers (Prio).Installed then
         raise Not_Scheduler_At_Priority;
      end if;

      --  Get appscheduler attribute
      Check_NZ (TO_APPSCHD.Pthread_Getappscheduler (To_MaRTE_Task_Id (Tid),
                                                    Sched'Access));
      if Sched /= Associated_Schedulers (Prio).Sched_Task_Id then
         raise Task_Not_Attached_To_Scheduler;
      end if;

      --  Get current scheduling parameters
      Check_NZ (TO.Pthread_Getschedparam (To_MaRTE_Task_Id (Tid),
                                          Policy'Access,
                                          Sched_Param'Access));
      --  Change policy
      Check_NZ (TO.Pthread_Setschedparam (To_MaRTE_Task_Id (Tid),
                                          Kernel.SCHED_FIFO,
                                          Sched_Param'Access));
   end Detach_Task_From_Scheduler;

   ------------------------------
   ----  Generate Abort Event  --
   ------------------------------
   procedure Generate_Abort_Event_For_Scheduler
     (Aborted_Task : in Ada.Task_Identification.Task_Id;
      Prio         : in System.Any_Priority) is
      function Ada_Task_Id_To_Int is
         new Unchecked_Conversion (Ada.Task_Identification.Task_Id, Int);
      Sched : aliased Kernel.Task_Id;
   begin
      if not Associated_Schedulers (Prio).Installed then
         raise Not_Scheduler_At_Priority;
      end if;

      --  Get appscheduler attribute
      Check_NZ (TO_APPSCHD.Pthread_Getappscheduler
                                           (To_MaRTE_Task_Id (Aborted_Task),
                                            Sched'Access));
      if Sched /= Associated_Schedulers (Prio).Sched_Task_Id then
         raise Task_Not_Attached_To_Scheduler;
      end if;

      if (SGNL_POSIX.Sigqueue (0, Associated_Schedulers (Prio).Signo,
                               Ada_Task_Id_To_Int (Aborted_Task)) /= 0)
      then
         raise Not_Enough_Resources_To_Send_Signal;
      end if;
   end Generate_Abort_Event_For_Scheduler;


   ----------------------------------------------------------------------------
   ----  Body of the Generic Scheduler Task  ----------------------------------
   ----------------------------------------------------------------------------
   task body Application_Scheduler_Task_Type is

      procedure Reset_Scheduling_Actions (Actions : in out Scheduling_Actions);
      procedure Reset_Scheduling_Actions
        (Actions : in out Scheduling_Actions) is
      begin
         Actions.MaRTE_Actions.Last := 0;
         Actions.Timeout_Active := False;
         Actions.CPU_Timers_Count := 0;
      end Reset_Scheduling_Actions;
      pragma Inline_Always (Reset_Scheduling_Actions);

      function Int_To_Ada_Task_Id is
         new Unchecked_Conversion (Int, Ada.Task_Identification.Task_Id);
      type Message_To_Scheduler_Ac is access Message_To_Scheduler'Class;
      function Address_To_Msg_Access is
        new Unchecked_Conversion (System.Address, Message_To_Scheduler_Ac);

      type Reply_From_Scheduler_Ac is access Reply_From_Scheduler'Class;
      function To_Reply_From_Scheduler_Ac is
         new Unchecked_Conversion (APPSCHD_DATA.Scheduler_Reply_Ac,
                                   Reply_From_Scheduler_Ac);
      Actions       : Scheduling_Actions;
      Error_Code    : Kernel.Error_Code;
      Signal_Waited : aliased Kernel.Signals.Signal_Set;
      Event_Ac      : APPSCHD_DATA.AppSched_Event_Ac;
      Flags         : Integer;
      Ret           : Integer;
      use type Kernel.Signals.Cause_Of_Signal;
   begin
      --  Become an application scheduler task
      SCHD.Self.AppScheduler := AppSched_Data_Ac;

      --  configure signal set with the only signal waited
      Check_NZ (SGNL_POSIX.Sig_Empty_Set (Signal_Waited'Access));
      Check_NZ (SGNL_POSIX.Sig_Add_Set (Signal_Waited'Access,
                                        Associated_Schedulers (Prio).Signo));

      --  Initialize Scheduler
      Init (Sched.all);

      --  Initialize actions
      Actions.MaRTE_Actions.Initialized := TO_APPSCHD.INITIALIZED;
      Reset_Scheduling_Actions (Actions);

      --  Processing events loop
      loop
         begin
            --  Execute Scheduling Actions
            Kernel.Enter_Critic_Section (Flags);
            TO_APPSCHD.Execute_Actions (Actions.MaRTE_Actions, Error_Code);
            Kernel.Leave_Critic_Section (Flags);

            --  Reset actions
            --  Reset_Scheduling_Actions (Actions);
            Actions.MaRTE_Actions.Last := 0;
            Actions.Timeout_Active := False;

            if Error_Code /= Kernel.NO_ERROR then
               --  Process error
               --
               if Error_Code = Kernel.INVALID_ARGUMENT then
                  Error (Sched.all,
                         Ada.Task_Identification.Null_Task_Id,

                         Ada.Application_Scheduling.Invalid_Action_For_Task,
                         Actions);
               else
                  Error (Sched.all,
                         Ada.Task_Identification.Null_Task_Id,

                         Ada.Application_Scheduling.Invalid_Action_For_Task,
                         Actions);
               end if;
            else
               --  Get a Scheduling Event or wait for it
               Kernel.Enter_Critic_Section (Flags);
               if Actions.Timeout_Active then
                  TO_APPSCHD.Get_Event
                    (Event_Ac,
                     Signal_Waited'Unrestricted_Access,
                     HWI.Duration_To_HWTime
                     (Ada.Real_Time.Delays.To_Duration (Actions.Timeout)));
               else
                  TO_APPSCHD.Get_Event (Event_Ac,
                                        Signal_Waited'Unrestricted_Access);
               end if;
               Kernel.Leave_Critic_Section (Flags);

--               Kernel_Console.Put
--                 (" --AppSchd:" &
--                  APPSCHD_DATA.Event_Code_T'Image (Event_Ac.Event_Code) &
--                  "--  ");

               --  Execute scheduler primitive depending on the error's
               --  type
               case Event_Ac.Event_Code is
                  when APPSCHD_DATA.APPSCHED_NEW =>
                     New_Task (Sched.all,
                               To_Ada_Task_Id (Event_Ac.T),
                               Actions);
                  when APPSCHD_DATA.APPSCHED_TERMINATE =>
                     Terminate_Task (Sched.all,
                                     To_Ada_Task_Id (Event_Ac.T),
                                     Actions);
                  when APPSCHD_DATA.APPSCHED_READY =>
                     Ready (Sched.all,
                            To_Ada_Task_Id (Event_Ac.T),
                            Actions);
                  when APPSCHD_DATA.APPSCHED_BLOCK =>
                     Block (Sched.all,
                            To_Ada_Task_Id (Event_Ac.T),
                            Actions);
                  when APPSCHD_DATA.APPSCHED_YIELD =>
                     Yield (Sched.all,
                            To_Ada_Task_Id (Event_Ac.T),
                            Actions);
                  when APPSCHD_DATA.APPSCHED_SIGNAL =>
                     if Event_Ac.Siginfo.Code = Kernel.Signals.SI_QUEUE then
                        Abort_Task
                          (Sched.all,
                           Int_To_Ada_Task_Id (Event_Ac.Siginfo.Value),
                           Actions);
                     else
                        null;


                     end if;

                  when APPSCHD_DATA.APPSCHED_CHANGE_SCHED_PARAM =>
                     Change_Sched_Param (Sched.all,
                                         To_Ada_Task_Id (Event_Ac.T),
                                         Actions);
                  when APPSCHD_DATA.APPSCHED_EXPLICIT_CALL_WITH_DATA =>

                     APPSCHD_DATA.UC_To_AppScheduler_Data_Ac
                       (SCHD.Self.AppScheduler).Reply_Size := 0;
                     --  >>
                     Explicit_Call
                       (Sched.all,
                        To_Ada_Task_Id (Event_Ac.T),
                        Address_To_Msg_Access (Event_Ac.Info),
                        To_Reply_From_Scheduler_Ac
                                     (APPSCHD_DATA.UC_To_Sched_App_Data_Ac
                                      (Event_Ac.T.Sched_App).Scheduler_Reply),
                        Actions);
                     Ret := TO_APPSCHD.POSIX_Appschedattr_Setreplyinfo (null,
                                                                        0);
                     pragma Assert (Ret = 0);
                  when APPSCHD_DATA.APPSCHED_TIMEOUT =>
                     Timeout (Sched.all,
                              Actions);
                  when APPSCHD_DATA.APPSCHED_PRIORITY_INHERIT =>
                     Priority_Inherit
                       (Sched          => Sched.all,
                        Tid            => To_Ada_Task_Id (Event_Ac.T),
                        From_Tid       => Ada.Task_Identification.Null_Task_Id,
                        Inherited_Prio =>
                          System.Any_Priority (Event_Ac.Sched_Priority),
                        Actions        => Actions);
                  when APPSCHD_DATA.APPSCHED_PRIORITY_UNINHERIT =>
                     Priority_Uninherit
                       (Sched            => Sched.all,
                        Tid              => To_Ada_Task_Id (Event_Ac.T),
                        From_Tid         =>
                          Ada.Task_Identification.Null_Task_Id,
                        Uninherited_Prio =>
                          System.Any_Priority (Event_Ac.Sched_Priority),
                        Actions          => Actions);
                  when APPSCHD_DATA.APPSCHED_TASK_NOTIFICATION =>
                     Task_Notification (Sched.all,
                                        To_Ada_Task_Id (Event_Ac.T),
                                        Actions);

                  when APPSCHD_DATA.APPSCHED_EXPLICIT_CALL |
                    APPSCHD_DATA.APPSCHED_INIT_MUTEX       |
                    APPSCHD_DATA.APPSCHED_DESTROY_MUTEX    |
                    APPSCHD_DATA.APPSCHED_LOCK_MUTEX       |
                    APPSCHD_DATA.APPSCHED_TRY_LOCK_MUTEX   |
                    APPSCHD_DATA.APPSCHED_UNLOCK_MUTEX     |
                    APPSCHD_DATA.APPSCHED_BLOCK_AT_MUTEX   |
                    APPSCHD_DATA.APPSCHED_CHANGE_SCHED_PARAM_MUTEX =>
                     --  Events not used in Ada implementation
                     null;
               end case;

               --  Event processed, then release it.
               APPSCHD_DATA.Release_AppSched_Event (Event_Ac);
            end if; --  if Error /= Kernel.NO_ERROR
         exception
            when E : others =>
               null;
               pragma Debug
                 (Kernel_Console.Error (Ada.Exceptions.Exception_Name (E)));
               pragma Debug
                 (Kernel_Console.Error (" " &
                                        Ada.Exceptions.Exception_Message (E)));
         end;
      end loop; --  Processing events loop

   end Application_Scheduler_Task_Type;

end Ada.Application_Scheduling.Support;
