------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.61B   070119
--
--    'A d a . A p p l i c a t i o n _ S c h e d u l i n g . S u p p o r t'
--
--                                  Spec
--
--
--  File 'a-apscsu.ads'                                                By MAR
--                                                                    and MGH
--
--
--  Suport for runtime and compiler implementation of
--  Application-Defined Scheduling.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with System;
with Ada.Task_Identification;

package Ada.Application_Scheduling.Support is

   ----------------------------------------------------------------------------
   --  Manage Schedulers and Appscheduled Tasks  ------------------------------
   ----------------------------------------------------------------------------
   procedure Install_Scheduler
     (Sched      : access Scheduler'Class;
      High_Prio  : in     System.Any_Priority;
      Low_Prio   : in     System.Any_Priority);
   --  Creates an application scheduler at the desired priority
   --  range.
   --
   --  A relationship between the scheduler and the priority range is
   --  established.


   Reached_Maximum_Number_Of_Application_Schedulers : exception;
   --  Raised by 'Install_Scheduler' when it is impossible to create
   --  the appscheduler since the maximum number of this kind of tasks
   --  has already been reached.


   function Scheduler_Installed (Prio : in System.Any_Priority)
                                return Boolean;
   --  Returns True if there is an user-defined scheduler installed
   --  at priority level 'Prio'


   procedure Attach_Task_To_Scheduler
     (Tid  : in Ada.Task_Identification.Task_Id;
      Prio : in System.Any_Priority);
   --  Attach the task 'Tid' with the application scheduler previously
   --  installed (using 'Install_Scheduler') at priority 'Prio'.
   --
   --  Assumes base priority of 'Tid' equal to 'Prio'.

   Task_Already_Attached_To_Appscheduler : exception;
   --  Raised by 'Attach_Task_To_Scheduler' when 'Tid' is attached to
   --  another appscheduler at the moment of the call (it should have been
   --  called 'Detach_Task_From_Scheduler' peviously)

   procedure Detach_Task_From_Scheduler
     (Tid  : in Ada.Task_Identification.Task_Id;
      Prio : in System.Any_Priority);
   --  Detach a task from the scheduler at priority 'Prio'.
   --
   --  Assumes base priority of 'Tid' equal to 'Prio' and 'Tid' has
   --  been attached previously to the scheduler using
   --  'Attach_Task_To_Scheduler'.
   --
   --  IMPORTANT: in order to change a task from one application
   --  scheduler to another it is necessary to call this procedure
   --  before calling 'Attach_Task_To_Scheduler' for the new
   --  scheduler.


   ---------------------------------------------------------------------------
   --  Generate Abort Event  -------------------------------------------------
   ---------------------------------------------------------------------------
   procedure Generate_Abort_Event_For_Scheduler
     (Aborted_Task : in Ada.Task_Identification.Task_Id;
      Prio         : in System.Any_Priority);
   --  Generates the "Abort_Task" event for 'Scheduler'. It is internally
   --  implemented sending a signal to the scheduler task.


   Not_Enough_Resources_To_Send_Signal : exception;
   --  Raised by 'Generate_Abort_Event_For_Scheduler' when MaRTE has
   --  not enough resources to send the signal.

   Task_Not_Attached_To_Scheduler : exception;
   --  Raised by 'Detach_Task_From_Scheduler' and
   --  'Generate_Abort_Event_For_Scheduler' when the task is not attached
   --  with the scheduler at priority 'Prio'.

   Not_Scheduler_At_Priority : exception;
   --  Raised by 'Attach_Task_To_Scheduler', 'Detach_Task_From_Scheduler'
   --  and 'Generate_Abort_Event_For_Scheduler' there is no scheduler
   --  installed at priority level 'Prio'.

   MaRTE_AppSched_Error : exception;
   --  Raised by 'Attach_Task_To_Scheduler',
   --  'Detach_Task_From_Scheduler' and
   --  'Generate_Abort_Event_For_Scheduler' when the
   --  application-scheduling implementation of MaRTE detects some
   --  error situation not included by the former exceptions.

end Ada.Application_Scheduling.Support;
