------------------------------------------------------------------------------
--                     FAT_Driver_Functions (body)                          --
------------------------------------------------------------------------------
--  This module provides a layer over FAT module in order to install it on  --
--  MaRTE OS, and be able to use a POSIX_IO interface. You have to install  --
--  it like any other driver and create the files (each file has the same   --
--  major number and a different minor number) in kernel-devices_table.ads  --
--  How to use it:  test_fat_driver.adb                                     --
--                                            daniel.sangorrin_at_gmail.com --
------------------------------------------------------------------------------
--  For unchecked conversions in Specific_Data
with Ada.Unchecked_Conversion;
with System;
--  for output messages
with Basic_Console_Io; use Basic_Console_Io;
--  for basic types
with Basic_Integer_Types; use Basic_Integer_Types;

package body Fat_Driver_Functions is

   ------------------------------------------------------------------
   -- Stuff for Specific_Data of each file (Entry and RW position) --
   ------------------------------------------------------------------

   type FAT_Specific_Data is record
      Ent : aliased FAT_Entry;
   end record;

   type FAT_Specific_Data_Ac is access all FAT_Specific_Data;
   function To_FAT_Specific_Data_Ac is new Ada.Unchecked_Conversion (Integer, FAT_Specific_Data_Ac);

   function FAT_Malloc (Size : in Size_T) return Integer;
   pragma Import (C, FAT_Malloc, "malloc");

   ------------
   -- Create --
   ------------
   Not_Initialized : boolean := true;

   function Create return DM.Int is
      Part : aliased IDE.Partitions.Partition;
   begin
      if Not_Initialized then
         --  Initialize the disk
         if IDE.Init (Disk) /= IDE.OK then
            return -1;
         end if;
         --  Get info of the partition we are going to be mount on
         if IDE.Partitions.Get_Partition_Info (Disk, Part_Num, Part'Access) /= IDE.OK then
            return -1;
         end if;
         --  Fill the filesystem info
         if FAT.Fat_Create (Part) /= 0 then
            return -1;
         end if;
         Not_Initialized := False;
         return 0;
      end if;
      return 0;
   end Create;

   ----------
   -- Open --
   ----------

   function Open (Fd   : in DM.File_Descriptor;
                  Mode : in DM.File_Access_Mode) return DM.Int
   is
      use type DM.File_Access_Mode;
      Ret : DM.Int;
      File_Name : DM.FSDT.Path;
      Invalid_Fd : Boolean;
      Data : Integer := FAT_Malloc (Size_T (FAT_Specific_Data'Size));
   begin
      DM.Get_Path_Of_Fd (Fd, File_Name, Invalid_Fd);
      if Invalid_Fd then
         return -1;
      else
         Ret := FAT.FAT_Open (File_Name,
                              To_FAT_Specific_Data_Ac (Data).Ent'Access,
                           (Mode and DM.File_Access_Mode (DM.O_CREAT)) = DM.O_CREAT);
         DM.Set_Specific_Data (Fd, Data, Invalid_Fd);
         return Ret;
      end if;
   end Open;

   ----------
   -- Read --
   ----------

   function Read (Fd         : in DM.File_Descriptor;
                  Buffer_Ptr : in DM.Buffer_Ac;
                  Bytes      : in DM.Buffer_Length) return DM.Ssize_T
   is
      use type DM.Buffer_Length;
      Ret : DM.Int;
      Invalid_Fd : Boolean;
      Data : Integer;
   begin
      DM.Get_Specific_Data (Fd, Data, Invalid_Fd);
      if Invalid_Fd then
         return -1;
      else
         Ret := FAT.FAT_Read (To_FAT_Specific_Data_Ac (Data).Ent'access,
                             Buffer_Ptr,
                             Bytes);
         return DM.Ssize_T (Ret);
      end if;
   end Read;

   -----------
   -- Write --
   -----------

   function Write (Fd         : in DM.File_Descriptor;
                   Buffer_Ptr : in DM.Buffer_Ac;
                   Bytes      : in DM.Buffer_Length) return DM.Ssize_T
   is
      use type DM.Buffer_Length;
      Ret : DM.Int;
      Invalid_Fd : Boolean;
      Data : Integer;
   begin
      DM.Get_Specific_Data (Fd, Data, Invalid_Fd);
      if Invalid_Fd then
         return -1;
      else
         Ret := FAT.FAT_Write (To_FAT_Specific_Data_Ac (Data).Ent'Access,
                               Buffer_Ptr,
                               Bytes);
         return DM.Ssize_T (Ret);
      end if;
   end Write;

   ------------
   -- Delete --
   ------------

   function Delete (Fd : in DM.File_Descriptor) return DM.Int is
      Invalid_Fd : Boolean;
      Data : Integer;
   begin
      DM.Get_Specific_Data (Fd, Data, Invalid_Fd);
      if Invalid_Fd then
         return -1;
      else
         return FAT.FAT_Delete (To_FAT_Specific_Data_Ac (Data).Ent);
      end if;
   end Delete;

   -----------
   -- Lseek --
   -----------

   function Lseek (Fd     : in DM.File_Descriptor;
                   Offset : in DM.Off_t;
                   Whence : in DM.Int) return FSDT.Off_t
   is
      Invalid_Fd : Boolean;
      Data : Integer;
   begin
      DM.Get_Specific_Data (Fd, Data, Invalid_Fd);
      if Invalid_Fd then
         return -1;
      else
         return FAT.FAT_Seek (To_FAT_Specific_Data_Ac (Data).Ent'Access,
                              Offset,
                              Whence);
      end if;
   end Lseek;

end Fat_Driver_Functions;