----------------------------------------------------------------------------
--  ---------------------           RT-EP           ------------------------
----------------------------------------------------------------------------
--                         Copyright (C) 2003-2005
--                     Universidad de Cantabria, SPAIN
--                        http://www.ctr.unican.es/
--
--    This program is free software; you can redistribute it and/or
--    modify it under the terms of the GNU General Public
--    License as published by the Free Software Foundation; either
--    version 2 of the License, or (at your option) any later version.
--
--    This program is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
--    General Public License for more details.
--
--    You should have received a copy of the GNU General Public
--    License along with this program; if not, write to the
--    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
--    Boston, MA 02111-1307, USA.
--
--    As a special exception, if other files instantiate generics from
--    this unit, or you link this unit with other files to produce an
--    executable, this  unit  does not  by itself cause  the resulting
--    executable to be covered by the GNU General Public License. This
--    exception does not however invalidate any other reasons why the
--    executable file  might be covered by the  GNU Public License.
--  ------------------------------------------------------------------------
--                                                     RT-EP /Ada95 {V1.1}
--
--  'r t e p p r o t o c o c h a n n e l s p r i o r i t y q u e u e . a d s'
--
--                                     Ada
--
--
--  File 'rtepprotocochannelspriorityqueue.ads'                    By Chema.
--                                                       Jose Maria Martinez
--                                                       <chema@gmx.net>
--
--  Priority FIFO queues. All the code except from the FIFO behaviour is
--  borrowed from Michael Gonzalez Harbour implementation (see headers below)
--
--
--
--
-----------------------------------------------------------------------------

-------------------------------------------------------------------------------
--                            PACKAGE PRIORITY_QUEUE
--                            ======================
--
--  This package defines a priority queue object and the operations that can be
--  performed on this object:
--
--     Empty and Full queue functions.
--     Enqueue : insert in the queue, in priority order.
--     Dequeue : extract the highest priority element from the queue. No
--               particular order is imposed on equal priority elements.
--     Dequeue_Middle : extract the queue element that matches the supplied
--                      argument.
--     Dequeue_Last : extract the lowest priority element from the queue. No
--                    particular order is imposed on equal priority elements.
--     Read_First : return the value of the highest priority element in the
--                  queue, without extracting it from the queue.
--
--  It is written as a generic package, in which the generic parameters are:
--
--     Size : Maximum number of elements in the queue
--     Element : type of the queue elements
--     Priority : type of the priority used for ordering the queue elements
--     ">" : Function used to order the priorities.
--     "=" : Used by dequeue_middle to compare the queue elements with
--           the element supplied by the caller.
-------------------------------------------------------------------------------
generic
   type FIFO_Index is mod <>; -- type FIFO_Index is mod Max_Elem_Same_Prio
   type Element is private;
   type Priority is range <>;
   with function ">" (Left,Right : in Priority) return Boolean;
package Priority_FIFO_Queues is

   --  The Priority_Storage_Full is risen when there is more same priority
   --  elements in the queue than the configured through FIFO_Index
   Priority_Storage_Full, Empty_Queue, Full_Queue : exception;

   type Prio_Queue is private;
   --  returns True if Empty;
   function Empty (Q : in Prio_Queue) return Boolean;

   function Full (Q : in Prio_Queue) return Boolean;

   --  Will raise Full_Queue.
   procedure Enqueue (Q : in out Prio_Queue;
                      E : in Element;
                      P : in Priority);

   --  Will raise Empty_Queue.
   procedure Dequeue (Q : in out Prio_Queue;
                      E : out Element;
                      P : out Priority);

   --  If threre is no elements in the queue, null is returned in E.
   procedure Read_First (Q : in out Prio_Queue;
                         E : out Element;
                         P : out Priority);

private

   Priority_Range : constant Integer := Integer (Priority'Last);

   Max_Elem_Same_Prio : constant Integer := Integer (FIFO_Index'Last) + 1;

   subtype Queued_Elements is Integer range
     0 .. (Max_Elem_Same_Prio*Priority_Range);

   subtype Prio_Index is Integer range 1 .. Priority_Range;

   subtype Queue_Size_Type is Integer range 0 .. Priority_Range;

   type Cell is record
      E   : Element;
      Pri : Priority;
   end record;

   type FIFO_Cell_Array is array (FIFO_Index) of Cell;

   type FIFO_Cell is record
      C : FIFO_Cell_Array;
      First : FIFO_Index := 0;
      Last : FIFO_Index := 0;
      --  The dirty state becomes true when there is an element in
      --  the heap within this priority. And false when there isn't.
      Dirty : Boolean := False;
      --  True if there are any elements in the stack. False if threr isn't
      Any_Element : Boolean := False;
   end record;

   type Queue_Cell_Array is array (Prio_Index) of Cell;
   type FIFO_FIFO_Cell_Array is array (Prio_Index) of FIFO_Cell;

   type Prio_Queue is record
      Queue : Queue_Cell_Array;
      FIFO : FIFO_FIFO_Cell_Array;
      Queue_Length : Queue_Size_Type := 0;
      Total_Length : Queued_Elements := 0;
   end record;

end Priority_FIFO_Queues;
